@@ -3,28 +3,15 @@ package Tk::More;
 use strict;
 use vars qw($VERSION @ISA);
 
-$VERSION = substr q$Revision: 1.8 $, 10;
+$VERSION = substr(q$Revision: 2.7 $, 10) . "";
 
+use Tk qw(Ev);
 use Tk::Derived;
 use Tk::Frame;
 @ISA = qw(Tk::Derived Tk::Frame);
 
 Construct Tk::Widget 'More';
 
-#sub ClassInit {
-#    my ($class, $mw) = @_;
-#
-#    $class->SUPER::ClassInit($mw);
-#
-#    ## xxx: useless 'because it's bound to frame and not rotext :-(
-#    $mw->bind($class, '<Key-h>', [qw(xview scroll -1 units)]);
-#    $mw->bind($class, '<Key-l>', [qw(xview scroll  1 units)]);
-#    $mw->bind($class, '<Key-k>', [qw(yview scroll -1 units)]);
-#    $mw->bind($class, '<Key-j>', [qw(yview scroll  1 units)]);
-#
-#    return $class;
-#};
-
 sub Populate {
     my ($cw, $args) = @_;
 
@@ -33,9 +20,21 @@ sub Populate {
 
     $cw->SUPER::Populate($args);
 
+    my $Entry = 'LabEntry';
+    my @Entry_args;
+    if (eval { die "Not yet";
+	       require Tk::HistEntry;
+	       Tk::HistEntry->VERSION(0.37);
+	       1;
+	   }) {
+	$Entry = 'HistEntry';
+    } else {
+	@Entry_args = (-labelPack=>[-side =>'left']);
+    }
+
     my $search;
-    my $e = $cw->LabEntry(
-		-labelPack=>[-side =>'left'],
+    my $e = $cw->$Entry(
+		@Entry_args,
 		-textvariable => \$search,
 		-relief => 'flat',
 		-state => 'disabled',
@@ -46,35 +45,52 @@ sub Populate {
     $cw->Advertise('text' => $t);
     $t->tagConfigure('search', -foreground => 'red');
 
+    # reorder bindings: private widget bindings first
+    $t->bindtags([$t, grep { $_ ne $t->PathName } $t->bindtags]);
+
     $t->bind('<Key-slash>',    [$cw, 'Search', 'Next']);
-# xxx forw/backw search should be recoded :-(
-#    $t->bind('<Key-question>', [$cw, 'Search', 'Prev']);
+    $t->bind('<Key-question>', [$cw, 'Search', 'Prev']);
     $t->bind('<Key-n>',        [$cw, 'ShowMatch', 'Next']);
     $t->bind('<Key-N>',        [$cw, 'ShowMatch', 'Prev']);
 
+    $t->bind('<Key-g>', $t->bind(ref($t),'<Control-Home>'));
     $t->bind('<Key-G>', $t->bind(ref($t),'<Control-End>'));
-    $t->bind('<Key-j>', ['yview', 'scroll',  1, 'units']);
-    $t->bind('<Key-k>', ['yview', 'scroll', -1, 'units']);
-    $t->bind('<Down>',  sub { $_[0]->yview('scroll',  1, 'units');
-			      Tk->break;
-			  });
-    $t->bind('<Up>',    sub { $_[0]->yview('scroll', -1, 'units');
-			      Tk->break;
-			  });
-    $t->bind('<Key-f>', $t->bind(ref($t),'<Next>'));
-    $t->bind('<Key-b>', $t->bind(ref($t),'<Prior>'));
+    $t->bind('<Home>',  $t->bind('<Key-g>'));
+    $t->bind('<End>',   $t->bind('<Key-G>'));
 
-    # Not documented (makes sense?)
-    $t->bind('<Key-l>', ['xview', 'scroll',  1, 'units']);
-    $t->bind('<Key-h>', ['xview', 'scroll', -1, 'units']);
-#    $t->bind('<Key-h>', $t->bind(ref($t),'<Left>'));
-#    $t->bind('<Key-l>', $t->bind(ref($t),'<Right>'));
-    $t->bind('<Return>', $t->bind(ref($t),'<Down>'));
-    $t->bind('<space>', $t->bind(ref($t),'<Next>'));
-    $t->bind('<Key-d>', $t->bind(ref($t),'<Next>'));  # xxx should be 1/2 screen
-    $t->bind('<Key-u>', $t->bind(ref($t),'<Prior>')); # xxx should be 1/2 screen
+    $t->bind('<Key-j>', [$cw, 'scroll', $t,  1, 'line']);
+    $t->bind('<Down>',  [$cw, 'scroll', $t,  1, 'line']);
+    $t->bind('<Key-k>', [$cw, 'scroll', $t, -1, 'line']);
+    $t->bind('<Up>',    [$cw, 'scroll', $t, -1, 'line']);
+
+    $t->bind('<Key-f>', [$cw, 'scroll', $t,  1, 'page']);
+    $t->bind('<Next>',  [$cw, 'scroll', $t,  1, 'page']);
+    $t->bind('<Key-b>', [$cw, 'scroll', $t, -1, 'page']);
+    $t->bind('<Prior>', [$cw, 'scroll', $t, -1, 'page']);
+
+    $t->bind('<Right>', [sub {
+		 return if ($_[1] =~ /(Alt|Meta)-/);
+		 $t->xview('scroll',  1, 'units'); Tk->break;
+	     }, Ev('s')]);
+    $t->bind('<Left>',  [sub {
+		 return if ($_[1] =~ /(Alt|Meta)-/);
+		 $t->xview('scroll', -1, 'units'); Tk->break;
+	     }, Ev('s')]);
+
+    $t->bind('<Return>', ['yview', 'scroll',  1, 'units']);
+    $t->bind('<Key-d>',  [$cw, 'scroll', $t,  1, 'halfpage']);
+    $t->bind('<Key-u>',  [$cw, 'scroll', $t, -1, 'halfpage']);
+
+    $t->bind('<Key-h>', sub { $cw->Callback(-helpcommand => $t) });
 
     $e->bind('<Return>',[$cw, 'SearchText']);
+    $e->bind('<Escape>',[$cw, 'SearchTextEscape']);
+
+    foreach my $mod (qw(Alt Meta)) {
+	foreach my $key (qw(n N g G j k f b d u h)) {
+	    $t->bind("<$mod-Key-$key>" => \&Tk::NoOp);
+	}
+    }
 
     $cw->Delegates('DEFAULT'   => $t,
 		   'Search'    => 'SELF',
@@ -87,6 +103,7 @@ sub Populate {
 		-padx          => [$t, qw(padX          Pad            5p)],
 		-pady          => [$t, qw(padY          Pad            5p)],
 		-searchcase    => ['PASSIVE', 'searchCase', 'SearchCase', 1],
+		-helpcommand   => ['CALLBACK', undef, undef, undef],
 		'DEFAULT'      => [$t]
 		);
 
@@ -100,48 +117,71 @@ sub Search {
     my $e = $cw->Subwidget('searchentry');
     $e->configure(-label => 'Search ' . ($direction eq 'Next'?'forward:':'backward:') );
     $e->configure(-relief=>'sunken',-state=>'normal');
+    $e->selectionRange(0, "end");
     $e->focus;
 }
 
 sub SearchText {
-    my ($cw) = @_;
+    my ($cw, %args) = @_;
     my($t, $e) = ($cw->Subwidget('text'), $cw->Subwidget('searchentry'));
-    unless ($cw->search_text($t, $e->get, 'search') ) {
-	$cw->bell;
+    $cw->{DIRECTION} = $args{-direction} if $args{-direction};
+    my $searchterm;
+    if (defined $args{-searchterm}) {
+	$searchterm = $args{-searchterm};
+	$ {$e->cget('-textvariable')} = $searchterm;
+    } else {
+	$e->historyAdd if ($e->can('historyAdd'));
+	$searchterm = $e->get;
+    }
+    unless ($cw->search_text($t, $searchterm, 'search') ) {
+	$cw->bell unless $args{-quiet};
     }
     $e->configure(-label=>'');
-    $t->see( '@1,0' );
-    # xxx better start at current pos as in more ???
-    $t->markSet('insert' ,'@1,0');
-    $cw->ShowMatch($cw->{DIRECTION});
+    $t->see('@0,0');
+    $cw->ShowMatch($cw->{DIRECTION}, -firsttime => 1) unless $args{-onlymatch};
+    $t->focus;
+    $e->configure(-relief=>'flat', -state=>'disabled');
+}
+
+sub SearchTextEscape {
+    my ($cw, %args) = @_;
+    my($t, $e) = ($cw->Subwidget('text'), $cw->Subwidget('searchentry'));
+    $e->configure(-label=>'');
     $t->focus;
     $e->configure(-relief=>'flat', -state=>'disabled');
 }
 
-# xxx when search changes from forward to backward (or vice versa)
-#     'insert' jumps from end to start of match (start to end)
-#     instead of the next (prev) match
 sub ShowMatch {
-    my ($cw, $method) = @_;
+    my ($cw, $method, %args) = @_;
+    my $firsttime = $args{-firsttime};
+
     my $t = $cw->Subwidget('text');
     if ($cw->{DIRECTION} ne 'Next') {
-	$method = 'Next' if $method eq 'Prev';	
-	$method = 'Prev' if $method eq 'Next';	
+	$method = 'Next' if $method eq 'Prev';
+	$method = 'Prev' if $method eq 'Next';
     }
-    my $cur = $t->index('insert');
+    my $cur = (($method eq 'Prev' && !$firsttime) ||
+	       ($method eq 'Next' &&  $firsttime)
+	       ? $t->index('@0,0')
+	       : $t->index('@0,'.$t->height));
     $method = "tag". $method . "range"; # $method: Next or Prev
     my @ins = $t->$method('search',$cur);
     unless (@ins) {
-	$cw->bell;
-	return;
+	# hack: Maybe the search was not performed yet? (e.g. after loading
+	# a new page but with the same search term)
+	my $e = $cw->Subwidget('searchentry');
+	if (!defined $ {$e->cget('-textvariable')}) {
+	    return;
+	}
+	$cw->SearchText(-searchterm => $ {$e->cget('-textvariable')},
+			-onlymatch => 1);
+	@ins = $t->$method('search',$cur);
+	return if !@ins;
     }
     @ins = reverse @ins unless $method eq 'tagNextrange';
-    $t->markSet('insert' ,$ins[1]);
-#XXX    my(@beforeVisible) = $t->yview;
     $t->see($ins[0]);
-#    my($currVisible)
     $ins[0];
-} 
+}
 
 # Load copied from TextUndo (xxx yy marks changes)
 sub Load
@@ -163,7 +203,8 @@ sub Load
   }
  else
   {
-   $text->BackTrace("Cannot open $file:$!");
+   $text->messageBox(-message => "Cannot open $file: $!\n");
+   die;
   }
 }
 
@@ -189,17 +230,60 @@ sub search_text {
     my $insert = $w->index('insert');
     my @search_args = ('-regexp');
     push @search_args, '-nocase' unless ($w->cget('-searchcase'));
-    while (1) {
-        $current = $w->search(@search_args, -count => \$length, '--', $string, $current, 'end');
-        last if not $current;
-	$found = 1;
-        $w->tag('add', $tag, $current, "$current + $length char");
-        $current = $w->index("$current + $length char");
+    eval {
+	while (1) {
+	    $current = $w->search(@search_args, -count => \$length, '--', $string, $current, 'end');
+	    last if not $current;
+	    $found = 1;
+	    $w->tag('add', $tag, $current, "$current + $length char");
+	    $current = $w->index("$current + $length char");
+	}
+	$w->markSet('insert', $insert);
+    };
+    if ($@) {
+	$w->messageBox(-icon => "error",
+		       -message => $@,
+		      );
     }
-    $w->markSet('insert', $insert);
     $found;
 } # end search_text
 
+sub scroll {
+    my($w,$t,$no,$unit) = @_;
+    if ($unit =~ /^line/) {
+	$t->yview('scroll', $no, 'units');
+    } else {
+	my($y1,$y2) = $t->yview;
+	my $amount;
+	if ($unit =~ /^halfpage/) {
+	    $amount = ($y2-$y1)/2;
+	} elsif ($unit =~ /^page/) {
+#  	    if ($no == -1) {
+#  		# loop until top-most line is invisible
+#  		my $inx = $t->index('@0,0');
+#  my $i=0;
+#  		while ($t->bbox($inx)) {
+#  		    $t->yviewScroll(-1,'units');
+#  		    last if ($i++>1000);
+#  		}
+#  		goto XXX;
+#  	    }
+	    $amount = ($y2-$y1);
+	} else {
+	    die "Unknown unit $unit";
+	}
+#warn "$y1 $y2 $amount";
+	$y1 += ($no * $amount);
+	if ($no > 0) {
+	    $y1 = 1.0 if ($y1 > 1.0);
+	} else {
+	    $y1 = 0.0 if ($y1 < 0.0);
+	}
+	$t->yviewMoveto($y1);
+    }
+    #XXX:
+    Tk->break;
+}
 
 
 #package Tk::More::Status;
@@ -225,37 +309,45 @@ Tk::More - a 'more' or 'less' like text widget
 =head1 DESCRIPTION
 
 B<Tk::More> is a readonly text widget with additional key bindings as
-found in UNI* command line tool C<more>.  As in C<more> an additional
-status/command line is added at the bottom.
+found in UNI* command line tools C<more> or C<less>. As in C<more> an
+additional status/command line is added at the bottom.
 
 =head1 ADDITIONAL BINDINGS
 
 =over 4
 
+=item Key-g
+
+goto beginning of file
+
 =item Key-G
 
 goto end of file
 
 =item Key-f
 
-like next key
+forward screen
 
 =item Key-b
 
-like prior key
+backward screen
 
 =item Key-k
 
-like up key
+up one line
 
 =item Key-j
 
-like down key
+down one line
 
 =item Key-/
 
 search forward
 
+=item Key-?
+
+search backward
+
 =item Key-n
 
 find next match
@@ -264,6 +356,14 @@ find next match
 
 find previous match
 
+=item Key-u
+
+up half screen
+
+=item Key-d
+
+down half screen
+
 =back
 
 =head1 BUGS
@@ -271,8 +371,6 @@ find previous match
 Besides that most of more bindings are not implemented. This bugs
 me most (high to low priority):
 
-* Reverse search missing
-
 * better status line implementation
 
 * Cursor movement: up/down move displayed area regardless where
@@ -280,19 +378,16 @@ me most (high to low priority):
 
 * add History, Load, Search (also as popup menu)
 
-* Key-u and Key-d should move 1/2 screen and not 90% of a screen
-
-* bad history impl.
-
 =head1 SEE ALSO
 
-L<Tk::ROText|Tk::ROText>
-more(1)
+L<Tk::ROText|Tk::ROText>, more(1), less(1)
 
 =head1 AUTHOR
 
 Achim Bohnet <F<ach@mpe.mpg.de>>
 
+Currently maintained by Slaven Rezic <F<slaven@rezic.de>>.
+
 Copyright (c) 1997-1998 Achim Bohnet. All rights reserved.  This program is
 free software; you can redistribute it and/or modify it under the same
 terms as Perl itself.
