@@ -1,9 +1,10 @@
 # -*- perl -*-
 
 #
+# $Id: Cache.pm,v 1.3 2003/01/29 11:57:19 eserte Exp $
 # Author: Slaven Rezic
 #
-# Copyright (C) 2002,2012 Slaven Rezic. All rights reserved.
+# Copyright (C) 2002 Slaven Rezic. All rights reserved.
 # This program is free software; you can redistribute it and/or
 # modify it under the same terms as Perl itself.
 #
@@ -18,14 +19,14 @@ use vars qw($VERSION $MAX_CACHE %CACHE);
 BEGIN {  # Make a DEBUG constant very first thing...
   if(defined &DEBUG) {
   } elsif(($ENV{'TKPODDEBUG'} || '') =~ m/^(\d+)/) { # untaint
-    my $debug = $1;
-    *DEBUG = sub () { $debug };
+    eval("sub DEBUG () {$1}");
+    die "WHAT? Couldn't eval-up a DEBUG constant!? $@" if $@;
   } else {
     *DEBUG = sub () {0};
   }
 }
 
-$VERSION = '5.02';
+$VERSION = sprintf("%d.%02d", q$Revision: 1.3 $ =~ /(\d+)\.(\d+)/);
 
 $MAX_CACHE = 20; # documents # XXX not yet used, LRU etc...
 
@@ -33,7 +34,7 @@ sub add_to_cache {
     my($w, $pod) = @_;
     $pod = $w->cget(-path) if !defined $pod;
     return if !defined $pod;
-    return if $CACHE{$pod}; # XXX check for recentness
+    return if exists $CACHE{$pod}; # XXX check for recentness
     DEBUG and warn "Add contents for $pod to cache.\n";
     $CACHE{$pod} = $w->dump_contents;
 }
@@ -41,7 +42,7 @@ sub add_to_cache {
 sub get_from_cache {
     my($w, $pod) = @_;
     $pod = $w->cget(-path) if !defined $pod;
-    return 0 if !$CACHE{$pod};
+    return 0 if !exists $CACHE{$pod};
     # XXX check for recentness
     $w->delete("1.0", "end");
     DEBUG and warn "Restore contents for $pod from cache.\n";
@@ -49,14 +50,6 @@ sub get_from_cache {
     1;
 }
 
-sub delete_from_cache {
-    my($w, $pod) = @_;
-    $pod = $w->cget(-path) if !defined $pod;
-    return if !defined $pod;
-    DEBUG and warn "Delete contents for $pod from cache.\n";
-    delete $CACHE{$pod};
-}
-
 sub clear_cache {
     %CACHE = ();
 }
@@ -64,14 +57,6 @@ sub clear_cache {
 sub dump_contents {
     my $w = shift;
     my @dump = $w->dump('-all', "1.0", "end");
-    if (@dump == 0) {
-	warn "Workaround strange bug under RedHat 8.0 --- try dump again...";	
-	@dump = $w->dump('-all', "1.0", "end");
-	if (@dump == 0) {
-	    warn "Giving up, cache disabled for current page";
-	    return undef;
-	}
-    }
     my %tags_def;
     foreach my $tag ($w->tagNames) {
 	# XXX check for used/existing tags missing
@@ -160,13 +145,3 @@ sub restore_contents {
 1;
 
 __END__
-
-=head1 NAME
-
-Tk::Pod::Cache - internal Tk-Pod module for cache control
-
-=head1 DESCRIPTION
-
-No user-servicable parts here.
-
-=cut
